/********************************************************************************
*                                                                               *
*                         B Z S t r e a m   C l a s s e s                       *
*                                                                               *
*********************************************************************************
* Copyright (C) 1999,2003 by Lyle Johnson. All Rights Reserved.                 *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
*********************************************************************************
* $Id: FXBzStream.cpp,v 1.9 2003/04/10 20:38:38 fox Exp $                       *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXBzStream.h"
#include <stdio.h>

#ifdef HAVE_BZ2LIB_H
#include <bzlib.h>

#define BLOCKSIZE100K 9         // Block size
#define VERBOSITY     0         // For tracing in bzip library
#define WORKFACTOR    0         // See bzip2 documentation
#define BZ_SAFE_SIZE(x) (128+(x*105)/100)


/*******************************************************************************/


namespace FX {


// Quickie to make an unsigned long out of two unsigned ints
inline unsigned long FXMAKELONG(unsigned int lo,unsigned int hi){
  return (hi<<32)+lo;
  }


// Initialize file stream
FXBzFileStream::FXBzFileStream(const FXObject* cont):FXStream(cont){
  file=NULL;
  bzfile=NULL;
  }


// Save to a file
void FXBzFileStream::saveItems(const void *buf,unsigned long n){
  int bzerror;
  if(dir!=FXStreamSave){ fxerror("FXBzFileStream::saveItems: wrong stream direction.\n"); }
  if(code==FXStreamOK){
    BZ2_bzWrite(&bzerror,(BZFILE*)bzfile,(void*)buf,n);
    if(bzerror!=BZ_OK){code=FXStreamFull;}
    }
  }


// Load from file
void FXBzFileStream::loadItems(void *buf,unsigned long n){
  int bzerror;
  if(dir!=FXStreamLoad){ fxerror("FXBzFileStream::loadItems: wrong stream direction.\n"); }
  if(code==FXStreamOK){
    BZ2_bzRead(&bzerror,(BZFILE*)bzfile,buf,n);
    if(bzerror!=BZ_OK){code=FXStreamEnd;}
    }
  }


// Try open file stream
FXbool FXBzFileStream::open(const char* filename,FXStreamDirection save_or_load){
  int bzerror;

  // Check for NULL filename
  if(!filename){ fxerror("FXBzFileStream::open: NULL filename argument.\n"); }

  // Stream should not yet be open
  if(dir!=FXStreamDead){ fxerror("FXBzFileStream::open: stream is already open.\n"); }

  // Open for read
  if(save_or_load==FXStreamLoad){
    file=fopen(filename,"rb");
    if(file==NULL){code=FXStreamNoRead;return FALSE;}
    bzfile=BZ2_bzReadOpen(&bzerror,(FILE*)file,0,VERBOSITY,NULL,0);
    if(bzerror!=BZ_OK){
      BZ2_bzReadClose(&bzerror,(BZFILE*)bzfile);
      code=FXStreamNoRead;
      return FALSE;
      }
    }

  // Open for write
  else{
    file=fopen(filename,"wb");
    if(file==NULL){code=FXStreamNoWrite;return FALSE;}
    bzfile=BZ2_bzWriteOpen(&bzerror,(FILE*)file,BLOCKSIZE100K,VERBOSITY,WORKFACTOR);
    if(bzerror!=BZ_OK){
      BZ2_bzWriteClose(&bzerror,(BZFILE*)bzfile,0,NULL,NULL);
      code=FXStreamNoWrite;
      return FALSE;
      }
    }

  // Do the generic book-keeping
  return FXStream::open(save_or_load);
  }


// Close file stream
FXbool FXBzFileStream::close(){
  int bzerror;
  if(bzfile){
    if(dir==FXStreamLoad)
      BZ2_bzReadClose(&bzerror,(BZFILE*)bzfile);
    else
      BZ2_bzWriteClose(&bzerror,(BZFILE*)bzfile,0,NULL,NULL);
    bzfile=NULL;
    }
  if(file){
    fclose((FILE*)file);
    file=NULL;
    }
  return FXStream::close();
  }


// Destructor
FXBzFileStream::~FXBzFileStream(){
  int bzerror;
  if(bzfile){
    if(dir==FXStreamLoad)
      BZ2_bzReadClose(&bzerror,(BZFILE*)bzfile);
    else
      BZ2_bzWriteClose(&bzerror,(BZFILE*)bzfile,0,NULL,NULL);
    bzfile=(void*)-1;
    }
  if(file){
    fclose((FILE*)file);
    file=(void*)-1;
    }
  }


/*******************************************************************************/


// Constructor
FXBzMemoryStream::FXBzMemoryStream(const FXObject *cont):FXStream(cont){
  FXCALLOC(&strm,bz_stream,1);
  bzpos=0;
  ptr=NULL;
  space=0;
  owns=FALSE;
  }


// Set available space
void FXBzMemoryStream::setSpace(unsigned long sp){
  if(!owns){ fxerror("FXBzMemoryStream::setSpace: cannot resize non-owned data buffer.\n"); }
  if(sp!=space){
    if(!FXRESIZE(&ptr,FXuchar,sp)){code=FXStreamAlloc;return;}
    space=sp;
    }
  }


// Save n bytes worth of data
void FXBzMemoryStream::saveItems(const void *buf,unsigned long n) {
  unsigned long bzSafeSize;
  if(dir!=FXStreamSave){ fxerror("FXBzMemoryStream::saveItems: wrong stream direction.\n"); }
  if(code==FXStreamOK){
    bzSafeSize = BZ_SAFE_SIZE(n)+bzpos;
    if(n>0){
      // Grow output buffer if necessary
      //if(bzpos+n>space){
      if(bzSafeSize>space){
        if(!owns){
	  code=FXStreamFull;
	  return;
	  }
//        setSpace(bzpos+n);
//        if(bzpos+n>space) return;
        setSpace(bzSafeSize);
        if(bzSafeSize>space) return;
        }
//      FXASSERT(bzpos+n<=space);
      FXASSERT(bzSafeSize<=space);
      bz_stream *bzs=(bz_stream*)strm;
      bzs->next_in=(char*)buf;
      bzs->avail_in=n;
      bzs->next_out=(char*)&ptr[bzpos];
      bzs->avail_out=space-bzpos;
      BZ2_bzCompress(bzs,BZ_RUN);
      FXASSERT(bzs->avail_in==0); // Confirm that we read all available data
      bzpos+=FXMAKELONG(bzs->total_out_lo32,bzs->total_out_hi32);
      }
    }
  }


// Request n bytes worth of data from the stream
void FXBzMemoryStream::loadItems(void *buf,unsigned long n){
  if(dir!=FXStreamLoad){ fxerror("FXBzMemoryStream::loadItems: wrong stream direction.\n"); }
  if(code==FXStreamOK){
    if(n>0){
      bz_stream *bzs=(bz_stream*)strm;
      bzs->next_in=(char*)&ptr[bzpos];
      bzs->avail_in=space-bzpos;
      bzs->next_out=(char*)buf;
      bzs->avail_out=n;
      unsigned long total_in_prev=FXMAKELONG(bzs->total_in_lo32,bzs->total_in_hi32);
      unsigned long total_out_prev=FXMAKELONG(bzs->total_out_lo32,bzs->total_out_hi32);
      BZ2_bzDecompress(bzs);
      bzpos=FXMAKELONG(bzs->total_in_lo32,bzs->total_in_hi32);
      FXASSERT(FXMAKELONG(bzs->total_out_lo32,bzs->total_out_hi32)-total_out_prev==n);
      }
    }
  }


// Open with indeterminate size
FXbool FXBzMemoryStream::open(FXuchar *data,FXStreamDirection save_or_load){
  int bzStatus;
  if(save_or_load==FXStreamSave)
    bzStatus=BZ2_bzCompressInit((bz_stream*)strm,BLOCKSIZE100K,VERBOSITY,WORKFACTOR);
  else
    bzStatus=BZ2_bzDecompressInit((bz_stream*)strm,VERBOSITY,0);
  bzpos=0;
  if(data){
    ptr=data;
    space=4294967295U;
    owns=FALSE;
    }
  else{
    FXCALLOC(&ptr,FXuchar,1);
    if(!ptr){
      code=FXStreamAlloc;
      return FALSE;
      }
    space=1;
    owns=TRUE;
    }
  return (bzStatus==BZ_OK) && FXStream::open(save_or_load);
  }


// Open with specified size
FXbool FXBzMemoryStream::open(FXuchar *data,unsigned long sp,FXStreamDirection save_or_load){
  int bzStatus;
  if(save_or_load==FXStreamSave)
    bzStatus=BZ2_bzCompressInit((bz_stream*)strm,BLOCKSIZE100K,VERBOSITY,WORKFACTOR);
  else
    bzStatus=BZ2_bzDecompressInit((bz_stream*)strm,VERBOSITY,0);
  bzpos=0;
  if(data){
    ptr=data;
    space=sp;
    owns=FALSE;
    }
  else{
    FXCALLOC(&ptr,FXuchar,sp);
    if(!ptr){
      code=FXStreamAlloc;
      return FALSE;
      }
    space=sp;
    owns=TRUE;
    }
  return (bzStatus==BZ_OK) && FXStream::open(save_or_load);
  }


// We're done; flush the compress/decompress buffers now
void FXBzMemoryStream::finish(){
  bz_stream *bzs=(bz_stream*)strm;
  if(dir==FXStreamLoad){
    int state=BZ_OK;
    state=BZ2_bzDecompress(bzs);
    BZ2_bzDecompressEnd(bzs);
    }
  else{
    int state=BZ_OK;
    bzs->next_in=NULL;
    bzs->avail_in=0;
    while(state!=BZ_STREAM_END){
      FXASSERT(space-bzpos>0);
      bzs->next_out=(char*)&ptr[bzpos];
      bzs->avail_out=space-bzpos;
      state=BZ2_bzCompress(bzs,BZ_FINISH);
      bzpos+=FXMAKELONG(bzs->total_out_lo32,bzs->total_out_hi32);
      }
    BZ2_bzCompressEnd(bzs);
    }
  }


// Take it away
void FXBzMemoryStream::takeBuffer(FXuchar*& buffer,unsigned long& sp) {
  finish();
  buffer=ptr;
  sp=space;
  ptr=NULL;
  space=0;
  owns=FALSE;
  }


// Give it back
void FXBzMemoryStream::giveBuffer(FXuchar* buffer,unsigned long sp) {
  if(buffer==NULL){ fxerror("FXMemoryStream::giveBuffer: NULL buffer argument.\n"); }
  if(owns){FXFREE(&ptr);}
  ptr=buffer;
  space=sp;
  owns=TRUE;
  }


// Close the stream
FXbool FXBzMemoryStream::close(){
  if(owns){FXFREE(&ptr);}
  ptr=NULL;
  space=0;
  owns=FALSE;
  return FXStream::close();
  }


// Destructor
FXBzMemoryStream::~FXBzMemoryStream(){
  FXFREE(&strm);
  if(owns){FXFREE(&ptr);}
  ptr=(FXuchar*)-1;
  }

}

#endif
